/***************************************************************************

  x11.c

  Common X11 routines for gb.qt and gb.gtk

  (c) 2000-2005 Benot Minisini <gambas@users.sourceforge.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __X11_C

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include "x11.h"

#define MAX_WINDOW_STATE 16

PUBLIC Atom X11_atom_net_wm_state;
PUBLIC Atom X11_atom_net_wm_state_above;
PUBLIC Atom X11_atom_net_wm_state_stays_on_top;
PUBLIC Atom X11_atom_net_wm_state_skip_taskbar;

static Display *_display;
static Window _root;

static bool _atom_init = FALSE;
static Atom _window_state[MAX_WINDOW_STATE];
static int _window_state_count = 0;


static void init_atoms(Display *dpy)
{
  if (_atom_init)
    return;
  
  X11_atom_net_wm_state = XInternAtom(dpy, "_NET_WM_STATE", True);
  X11_atom_net_wm_state_above = XInternAtom(dpy, "_NET_WM_STATE_ABOVE", True);
  X11_atom_net_wm_state_stays_on_top = XInternAtom(dpy, "_NET_WM_STATE_STAYS_ON_TOP", True);  
  X11_atom_net_wm_state_skip_taskbar = XInternAtom(dpy, "_NET_WM_STATE_SKIP_TASKBAR", True);
  
  _atom_init = TRUE;
}

static void load_window_state(Display *dpy, Window wid)
{
  Atom type;
  int format;
  unsigned long length, after;
  unsigned char *data;

  init_atoms(dpy);
    
  _window_state_count = 0;
  
  XGetWindowProperty(dpy, wid, X11_atom_net_wm_state, 0, 8,
    False, AnyPropertyType, &type, &format,
    &length, &after, &data);  
    
  if (length > 8)
    length = 8;
    
  _window_state_count = length;
  memcpy(_window_state, data, _window_state_count * sizeof(Atom));
  
  XFree(data);
}

static void save_window_state(Display *dpy, Window wid)
{
  XChangeProperty(dpy, wid, X11_atom_net_wm_state, XA_ATOM, 32, PropModeReplace,
      (unsigned char *)_window_state, _window_state_count);
}

static bool is_window_state(Atom state)
{
  int i;
  
  for (i = 0; i < _window_state_count; i++)
  {
    if (_window_state[i] == state)
      return TRUE;
  }
  
  return FALSE;
}

static void set_window_state(Atom state)
{
  if (is_window_state(state))
    return;
    
  if (_window_state_count == MAX_WINDOW_STATE)
  {
    fprintf(stderr, "X11: set_window_state: Too many properties in window\n");
    return;
  }
  
  _window_state[_window_state_count++] = state;
}
    
static void clear_window_state(Atom state)
{
  int i;

  for (i = 0; i < _window_state_count; i++)
  {
    if (_window_state[i] == state)
    {
      _window_state_count--;
    
      for (; i < _window_state_count; i++)
        _window_state[i] = _window_state[i + 1];
        
      return;
    }
  }
  
}


PUBLIC void X11_init(Display *display, Window root)
{
  _display = display;
  _root = root;
}


PUBLIC void X11_window_change_property(Window window, bool visible, Atom property, bool set)
{
  XEvent e;
  long mask = (SubstructureRedirectMask | SubstructureNotifyMask);

  if (visible)
  {
    e.xclient.type = ClientMessage;
    e.xclient.message_type = X11_atom_net_wm_state;
    e.xclient.display = _display;
    e.xclient.window = window;
    e.xclient.format = 32;
    e.xclient.data.l[0] = set ? 1 : 0;
    e.xclient.data.l[1] = property;
    e.xclient.data.l[2] = 0;
    e.xclient.data.l[3] = 0;
    e.xclient.data.l[4] = 0;
  
    XSendEvent(_display, _root, False, mask, &e);
  }
  else
  {
    load_window_state(_display, window);
    
    if (set)
      set_window_state(property);
    else
      clear_window_state(property);
    
    save_window_state(_display, window);
  }
}


PUBLIC bool X11_window_has_property(Window window, Atom property)
{
  load_window_state(_display, window);
  return is_window_state(property);
}

PUBLIC void X11_sync(void)
{
  XSync(_display, False);
}

PUBLIC void X11_window_save_properties(Window window)
{
  load_window_state(_display, window);
}

PUBLIC void X11_window_restore_properties(Window window)
{
  save_window_state(_display, window);
}

#define SYSTEM_TRAY_REQUEST_DOCK    0
#define SYSTEM_TRAY_BEGIN_MESSAGE   1
#define SYSTEM_TRAY_CANCEL_MESSAGE  2

#define OPCODE "_NET_SYSTEM_TRAY_OPCODE"

PUBLIC void X11_window_dock(Window window)
{
  Window xmanager=None;
  XClientMessageEvent ev;
  Atom OpCodeAtom;
  Screen *xscreen;
  char buf[256];
  Atom selection_atom;
  
  buf[0]=0;
  
  xscreen = DefaultScreenOfDisplay(_display);
  sprintf(buf,"_NET_SYSTEM_TRAY_S%d",XScreenNumberOfScreen(xscreen));
  selection_atom = XInternAtom(_display, buf, 0);
          
  XGrabServer(_display);

  xmanager = XGetSelectionOwner(_display, selection_atom);
  if (xmanager != None)
    XSelectInput(_display, xmanager, StructureNotifyMask);
          
  XUngrabServer(_display);
  XFlush(_display);
  
  /***********************************************
    Dock Tray Icon
  ************************************************/
  
  OpCodeAtom = XInternAtom(_display, OPCODE, 0);
  
  ev.type = ClientMessage;
  ev.window = xmanager;
  ev.message_type = OpCodeAtom;
  ev.format = 32;
  ev.data.l[0] = 0;
  ev.data.l[1] = SYSTEM_TRAY_REQUEST_DOCK;
  ev.data.l[2] = window;
  ev.data.l[3] = 0;
  ev.data.l[4] = 0;

  XSendEvent(_display, xmanager, 0, NoEventMask, (XEvent *)&ev);
  XSync(_display, 0);	
}

PUBLIC void X11_window_startup(Window window, int x, int y, int w, int h)
{
  XSizeHints s;
  
  s.flags = USPosition | PPosition | USSize | PSize;
  
  s.x = x;
  s.y = y;
  s.width = w;
  s.height = h;
  
  XSetWMNormalHints(_display, window, &s);
}
